import re
import json
import subprocess
import ipaddress
from flask import Flask, request, abort
from .gitlab_api import GitlabApi
from .helpers import Helpers

app = Flask(__name__)

whitelist_ip = None
repos = None

@app.route("/", methods=['GET', 'POST'], strict_slashes=False)
def index():
    if request.method == "GET":
        return 'OK'
    elif request.method == "POST":
        # Check the POST source
        if not whitelist_ip is None:
            # Store the IP address of the requester
            request_ip = ipaddress.ip_address(request.remote_addr)

            for block in [whitelist_ip]:
                if request_ip in ipaddress.ip_network(block):
                    break # remote_addr is within the accepted network range
            else:
                abort(403)

        if not len(request.data):
            abort(400)

        try:
            payload = json.loads(request.data)
        except:
            abort(400)

        # common for events
        if payload['object_kind'] in ['push', 'issue']:
            repo_meta = {
                'homepage': payload['repository']['homepage'],
            }
            repo = repos.get(repo_meta['homepage'], None)
            if not repo:
                return "Nothing to do for " + repo_meta['homepage']

            webhook_token = repo.get('webhook_token', None)
            private_token = repo.get('private_token', None)
        else:
            return "Unsupported object kind", 422

        if webhook_token:
            # validate against X-Gitlab-Token header
            request_token = request.headers.get('X-Gitlab-Token', None)
            if not request_token or request_token != webhook_token:
                abort(403)

        if payload['object_kind'] == "push":
            match = re.match(r"refs/heads/(?P<branch>.*)", payload['ref'])
            if match:
                repo_meta['branch'] = match.groupdict()['branch']
            else:
                return "Unable to determine pushed branch"

            push = repo.get("push", None)
            if push:
                branch = push.get(repo_meta['branch'], None)
                if not branch:
                    branch = repo['push'].get("other", None)
                if branch:
                    branch_actions = branch.get("actions", None)

                    if branch_actions:
                        for action in branch_actions:
                            try:
                                subp = subprocess.Popen(action, cwd=branch.get("path", "."), shell=True)
                                subp.wait()
                            except Exception as e:
                                print(e)
            return 'OK'

        if payload['object_kind'] == "issue":
            issue = repo.get("issue", None)
            if issue:
                # notification for new issue
                if issue.get("user_notify", None) and payload['object_attributes']['action'] == "open":
                    if not private_token:
                        abort(403)
                    gl = GitlabApi(repo_meta['homepage'], private_token)
                    notify = issue['user_notify']
                    description = payload['object_attributes']['description']
                    usernames = []
                    for n in notify:
                        username_match = re.match("^@[a-zA-Z0-9_.+-]+$", n)
                        if username_match:
                            # simple username
                            usernames.append(n)
                        else:
                            # try to pull the email from the issue body
                            # and derive the username from that
                            body_match = re.match(n, description)
                            if body_match:
                                email = body_match.group(1)
                                username = gl.lookup_username(email)
                                if username:
                                    usernames.append("@" + username)
                    # narrow down to unique names
                    usernames = list(set(usernames))
                    if len(usernames) > 0:
                        project_id = payload['object_attributes']['project_id']
                        issue_id = payload['object_attributes']['id']
                        gl.comment_on_issue(project_id, issue_id, "Automatic mention for %s" % (" and ".join(usernames)))

                # parse commit message and manage labels on issues
                if issue.get("labels"):
                    if not private_token:
                        abort(403)
                    gl = GitlabApi(repo_meta['homepage'], private_token)
                    helpers = Helpers()
                    project_id = payload['object_attributes']['project_id']
                    labels = helpers.get_label_names(gl.get_labels(project_id))
                    list_labels = helpers.get_list_labels(gl.get_boards(project_id))
                    for commit in payload['commits']:
                        parse_commit = helpers.parse_commit_labels(commit['message'])
                        for issue in parse_commit['issues']:
                            issue_labels = gl.get_issue(project_id, issue)
                            updated_labels = helpers.simplify_labels(issue_labels, parse_commit['label_ops'])
                            gl.set_issue_labels(project_id, issue, updated_labels)
            return 'OK'




import argparse
import json

from .__version__ import (
    __title__,
    __version__
)
from gwh import app
import gwh

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="python -m gwh", description="GitLab webhook handler")
    parser.add_argument("config", help="path to repos configuration")
    parser.add_argument("--version", action="version", version=__title__ + " " + __version__)
    parser.add_argument("--host", help="server host", default="0.0.0.0")
    parser.add_argument("-p", "--port", help="server port", type=int, default=8080)
    parser.add_argument("--allow", help="whitelist GitLab IP block")
    parser.add_argument("--debug", action="store_true", help="enable debug output")

    args = parser.parse_args()

    try:
        gwh.repos = json.loads(open(args.config, encoding='utf-8').read())
    except:
        print("Error opening repos file %s -- check file exists and is valid JSON" % args.config)
        raise

    gwh.whitelist_ip = args.allow

    app.debug = args.debug

    app.run(host=args.host, port=args.port)



import os
from setuptools import setup

setup(
    name="gwh",
    version="1.0.1",
    author='Andy Hebrank',
    author_email='ahebrank@gmail.com',
    packages=['gwh'],
    url='https://github.com/ahebrank/gitlab-webhook-handler',
    description='Webhook Handler for GitLab',
    license='Apache License, Version 2.0',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    install_requires=[
        'Flask>=1.0',
        'requests>=2.19.0'
    ]
)



import re

class Helpers:

    def get_list_labels(self, boards):
        # return label IDs from all project boards
        label_ids = []
        for board in boards:
            label_ids += [list_label['label']['name'] for list_label in board['lists']]
        return label_ids

    def get_label_names(self, labels):
        return [label['name'] for label in labels]

    def get_issue_labels(self, issue):
        return issue['labels']

    def parse_commit_labels(self, message, labels):
        # get the issue numbers
        issue_pat = re.compile('#(\d+)')
        iid_match = issue_pat.findall(message)
        
        # get the labels
        label_pat = re.compile('([\~\+\-])(' + '|'.join(labels) + ')')
        label_match = label_pat.findall(message)
        return {
            'issues': iid_match,
            'label_ops': label_match
        }

    def simplify_labels(self, existing, from_message, board_labels = None):
        labels = existing
        for ops in from_message['label_ops']:
            op = ops[0]
            name = ops[1].decode('utf-8')
            if op == '+':
                labels.append(name)
            if op == '-':
                labels.remove(name)
            if op == '~':
                if name in board_labels:
                    # remove all board labels
                    labels = [label for label in labels if label not in board_labels]
                    # set board label
                    labels.append(name)
        return labels
